import base64
import json
import re
from json import JSONDecodeError


class ExtendedMetaData:
    _base64_pattern = '(?:[A-Za-z0-9+/]{4})*(?:[A-Za-z0-9+/]{2}==|[A-Za-z0-9+/]{3}=)?=?'
    _metadata_pattern = f'^EMD: ({_base64_pattern})$'

    def __init__(self, data):
        self._data = data

    def value(self, field):
        try:
            return self._data[field]
        except KeyError:
            return None

    @staticmethod
    def create(raw_value):
        re_result = re.search(ExtendedMetaData._metadata_pattern, str(raw_value))

        if re_result is None:
            return None

        encoded_metadata = re_result.group(1)

        if encoded_metadata is None:
            return None

        try:
            meta_data = json.loads(base64.b64decode(encoded_metadata))
        except JSONDecodeError:
            return None

        return ExtendedMetaData(meta_data)



import unittest

from beetsplug.extendedmetadata import ExtendedMetaDataMatchQuery


class ExtendedMetaDataMatchQueryTest(unittest.TestCase):

    def setUp(self):
        # {"origin":"korea","language":["korean","english"],"vocal_gender":"male", "genre":"k-pop", "tag":["two words"], "test":"two words", "jpn_word":"けいおん!"}
        self.extended_metadata = 'EMD: eyJvcmlnaW4iOiJrb3JlYSIsImxhbmd1YWdlIjpbImtvcmVhbiIsImVuZ2xpc2giXSwidm9jYWxfZ2VuZGVyIjoibWFsZSIsICJnZW5yZSI6ImstcG9wIiwgInRhZyI6WyJ0d28gd29yZHMiXSwgInRlc3QiOiJ0d28gd29yZHMiLCAianBuX3dvcmQiOiLjgZHjgYTjgYrjgpMhIn0='
        self.sut = ExtendedMetaDataMatchQuery(None, None)

    # INVALID PATTERNS #
    def test_invalid_pattern(self):
        self.assertFalse(self.sut.value_match('language:,', self.extended_metadata))
        self.assertFalse(self.sut.value_match('language:,korean', self.extended_metadata))
        self.assertFalse(self.sut.value_match('language:korean,', self.extended_metadata))
        self.assertFalse(self.sut.value_match('language:korean,english,', self.extended_metadata))
        self.assertFalse(self.sut.value_match('language:!', self.extended_metadata))

    def test_empty_json(self):
        self.assertFalse(self.sut.value_match('origin:', '{}'))

    def test_empty_non_json(self):
        self.assertFalse(self.sut.value_match('origin:', 'korea'))

    def test_empty_pattern(self):
        self.assertTrue(self.sut.value_match('origin:', self.extended_metadata))

    # SINGLE VALUE #
    def test_valid_pattern(self):
        self.assertTrue(self.sut.value_match('origin:korea', self.extended_metadata))
        self.assertTrue(self.sut.value_match('genre:k-pop', self.extended_metadata))

    def test_different_case(self):
        self.assertTrue(self.sut.value_match('origin:Korea', self.extended_metadata))

    def test_unicode_characters(self):
        self.assertTrue(self.sut.value_match('jpn_word:けいおん!', self.extended_metadata))

    def test_negated(self):
        self.assertFalse(self.sut.value_match('origin:!korea', self.extended_metadata))
        self.assertTrue(self.sut.value_match('origin:!usa', self.extended_metadata))
        self.assertTrue(self.sut.value_match('tag:!vocaloid', self.extended_metadata))

    def test_spaces(self):
        self.assertTrue(self.sut.value_match('tag:two words', self.extended_metadata))
        self.assertTrue(self.sut.value_match('test:two words', self.extended_metadata))

    # MULTI VALUE #
    def test_array_negated(self):
        self.assertFalse(self.sut.value_match('language:!korean', self.extended_metadata))
        self.assertTrue(self.sut.value_match('language:!japanese', self.extended_metadata))
        self.assertTrue(self.sut.value_match('language:!korean,english', self.extended_metadata))
        self.assertFalse(self.sut.value_match('language:!korean,!english,', self.extended_metadata))

    def test_array_one_of_match(self):
        self.assertTrue(self.sut.value_match('language:korean,japanese', self.extended_metadata))
        self.assertTrue(self.sut.value_match('language:chinese,english', self.extended_metadata))
        self.assertFalse(self.sut.value_match('language:chinese,japanese', self.extended_metadata))

    def test_array_empty_pattern(self):
        self.assertTrue(self.sut.value_match('language:', self.extended_metadata))

    def test_non_existing_pattern(self):
        self.assertFalse(self.sut.value_match('origin:usa', self.extended_metadata))

    def test_array_existing_pattern(self):
        self.assertTrue(self.sut.value_match('language:korean', self.extended_metadata))

    def test_array_existing_pattern_different_case(self):
        self.assertTrue(self.sut.value_match('language:Korean', self.extended_metadata))

    def test_array_non_existing_pattern(self):
        self.assertFalse(self.sut.value_match('language:japanese', self.extended_metadata))




from beets.dbcore import FieldQuery
from beets.plugins import BeetsPlugin

from beetsplug.emd_metadata import ExtendedMetaData
from beetsplug.emd_query import create_query
from beetsplug.emd_audiofilefields import fields


class ExtendedMetaDataMatchQuery(FieldQuery):

    @classmethod
    def value_match(cls, pattern, val):
        meta_data = ExtendedMetaData.create(val)

        if meta_data is None:
            return False

        query = create_query(pattern)

        if query is None:
            return False

        return query.matches(meta_data)


class ExtendedMetaDataPlugin(BeetsPlugin):

    def __init__(self):
        super(ExtendedMetaDataPlugin, self).__init__()

        self.input_field = self.config['input_field'].get('comments')
        self.query_field = self.config['query_field'].get('x')

        if fields[self.input_field] is not None:
            extended_meta_data = fields[self.input_field]
            self.add_media_field(u'' + self.query_field, extended_meta_data)

    def queries(self):
        return {
            '.': ExtendedMetaDataMatchQuery
        }




import unittest

from beetsplug.extendedmetadata import ExtendedMetaDataMatchQuery


class ExtendedMetaDataMatchQueryTest(unittest.TestCase):

    def setUp(self):
        # {"origin":"korea","language":["korean","english"],"vocal_gender":"male", "genre":"k-pop", "tag":["two words"], "test":"two words", "jpn_word":"けいおん!"}
        self.extended_metadata = 'EMD: eyJvcmlnaW4iOiJrb3JlYSIsImxhbmd1YWdlIjpbImtvcmVhbiIsImVuZ2xpc2giXSwidm9jYWxfZ2VuZGVyIjoibWFsZSIsICJnZW5yZSI6ImstcG9wIiwgInRhZyI6WyJ0d28gd29yZHMiXSwgInRlc3QiOiJ0d28gd29yZHMiLCAianBuX3dvcmQiOiLjgZHjgYTjgYrjgpMhIn0='
        self.sut = ExtendedMetaDataMatchQuery(None, None)

    # SINGLE VALUE #
    def test_invalid_pattern(self):
        self.assertFalse(self.sut.value_match('language::^$^', self.extended_metadata))

    def test_empty_pattern(self):
        self.assertTrue(self.sut.value_match('origin::', self.extended_metadata))

    def test_existing_pattern(self):
        self.assertTrue(self.sut.value_match('origin::^.*ea$', self.extended_metadata))

    def test_non_existing_pattern(self):
        self.assertFalse(self.sut.value_match('origin::us.*', self.extended_metadata))

    # MULTI VALUE #
    def test_array_empty_pattern(self):
        self.assertTrue(self.sut.value_match('language::', self.extended_metadata))

    def test_array_existing_pattern(self):
        self.assertTrue(self.sut.value_match('language::^.*an$', self.extended_metadata))

    def test_array_non_existing_pattern(self):
        self.assertFalse(self.sut.value_match('language::ja.*', self.extended_metadata))
